from .wastuff import sync_service, async_service, wasbase

class AsyncWAS (wasbase.WASBase, async_service.AsyncService):
    pass
WAS = AsyncWAS

class SyncWAS (wasbase.WASBase, sync_service.SyncService):
    pass
        





import sys
from ..utility import make_pushables
from ..exceptions import HTTPError
from . import corequest, response
from .httpbase.task import DEFAULT_TIMEOUT
from skitai import was

class TaskBase (corequest):
    def __init__ (self, reqs, timeout = DEFAULT_TIMEOUT, **meta):
        assert isinstance (reqs, (list, tuple))
        self._timeout = timeout                
        self._reqs = reqs
        self.meta = meta

    def __getattr__ (self, name):
        try:
            return self.meta [name]
        except KeyError:
            raise AttributeError ("{} cannot found".format (name))    

class Tasks (TaskBase):
    def __init__ (self, reqs, timeout = DEFAULT_TIMEOUT, **meta):
        TaskBase.__init__ (self, reqs, timeout, **meta)
        self._results = []        

    def __iter__ (self):
        return iter (self._reqs)
    
    def __getitem__ (self, sliced):
        return self._reqs [sliced].dispatch ()
    
    def set_timeout (self, timeout):
        self._timeout = timeout

    def add (self, req):
        self._reqs.append (req)

    def merge (self, tasks):
        for req in tasks._reqs:
            self.add (req)        

    def then (self, func):
        return was.Futures (self._reqs, self._timeout, **self.meta).then (func)

    #------------------------------------------------------
    def cache (self, cache = 60, cache_if = (200,)):
        [r.cache (cache, cache_if) for r in self.results]

    def dispatch (self, cache = None, cache_if = (200,), timeout = None):
        return [req.dispatch (cache, cache_if, timeout or self._timeout) for req in self._reqs]
        
    def wait (self, timeout = None):
        return [req.wait (timeout or self._timeout) for req in self._reqs]
        
    def commit (self, timeout = None):
        [req.commit (timeout or self._timeout) for req in self._reqs] 
    
    def fetch (self, cache = None, cache_if = (200,)):
        return [req.fetch (cache, cache_if) for req in self._reqs]
        
    def one (self, cache = None, cache_if = (200,), timeout = None):
        return [req.one (cache, cache_if) for req in self._reqs]


class Mask (response, TaskBase):
    def __init__ (self, data, **meta):
        self._data = data
        self.meta = meta

    def commit (self):
        pass

    def fetch (self):
        return self._data
    
    def one (self):    
        if len (self._data) == 0:
            raise HTTPError ("404 Not Found")
        if len (self._data) != 1:
            raise HTTPError ("409 Conflict")
        if isinstance (self._data, dict):
            return self._data.popitem () [1]
        return self._data [0]


class CompletedTasks (response, Tasks):
    def __init__ (self, reqs, **meta):
        Tasks.__init__ (self, reqs, **meta)


class Futures (TaskBase):
    def __init__ (self, was, reqs, timeout = 10, **meta):
        TaskBase.__init__ (self, reqs, timeout, **meta)
        self._was = was        
        self._fulfilled = None
        self._responded = 0
            
    def then (self, func):        
        self._fulfilled = func
        for reqid, req in enumerate (self._reqs):
           req.set_callback (self._collect, reqid, self._timeout)
        return self
                 
    def _collect (self, res):
        self._responded += 1        
        if self._responded == len (self._reqs):
            if self._fulfilled:             
                self._respond ()
            else:
                self._was.response ("205 No Content", "")
                self._was.response.done ()
            
    def _respond (self):
        response = self._was.response         
        try:
            tasks = CompletedTasks (self._reqs, **self.meta)
            content = self._fulfilled (self._was, tasks)
            will_be_push = make_pushables (response, content)
            content = None
        except MemoryError:
            raise
        except HTTPError as e:
            response.start_response (e.status)
            content = response.build_error_template (e.explain, e.errno, was = self._was)
        except:            
            self._was.traceback ()
            response.start_response ("502 Bad Gateway")
            content = response.build_error_template (self._was.app.debug and sys.exc_info () or None, 0, was = self._was)            
       
        if content:
           will_be_push = make_pushables (response, content)
        
        if will_be_push is None:
            return
           
        for part in will_be_push:
            if len (will_be_push) == 1 and type (part) is bytes and len (response) == 0:
                response.update ("Content-Length", len (part))
            response.push (part)                
        response.done ()
    



#!/usr/bin/env python

import sys
import asyncore, asynchat
import re, socket, time, threading, os
from . import http_request, counter
from aquests.protocols.http import http_util, http_date
from aquests.athreads import threadlib
from skitai import lifetime
from rs4 import producers, compressors
from collections import deque
import signal
import ssl
import skitai
from hashlib import md5
from rs4.psutil.processutil import set_process_name
if os.name == "posix":
	import psutil
	CPUs = psutil.cpu_count()

PID = {}
ACTIVE_WORKERS = 0
SURVAIL = True
EXITCODE = 0
DEBUG = False

	
#-------------------------------------------------------------------
# server channel
#-------------------------------------------------------------------
class http_channel (asynchat.async_chat):	
	current_request = None
	channel_count = counter.counter ()
	closed = False
	is_rejected = False
	
	keep_alive = 30
	network_timeout = 30
	zombie_timeout = 30
	
	fifo_class = deque
	
	def __init__ (self, server, conn, addr):
		self.channel_number = http_channel.channel_count.inc ()
		self.request_counter = counter.counter()
		self.bytes_out = counter.counter()
		self.bytes_in = counter.counter()
		
		#asynchat.async_chat.__init__ (self, conn)
		self.ac_in_buffer = b''
		self.incoming = []
		self.producer_fifo = self.fifo_class ()
		asyncore.dispatcher.__init__(self, conn)
		
		self.server = server
		self.addr = addr		
		self.set_terminator (b'\r\n\r\n')
		self.in_buffer = b''
		self.creation_time = int (time.time())
		self.event_time = int (time.time())
		self.__history = []
		self.__sendlock = None
		self.producers_attend_to = []
		self.things_die_with = []		
		
	def use_sendlock (self):
		self.__sendlock  = threading.Lock ()
		self.initiate_send = self._initiate_send_ts
			
	def get_history (self):
		return self.__history
			
	def reject (self):
		self.is_rejected = True		
	
	def _initiate_send_ts (self):
		lock = self.__sendlock
		lock.acquire ()
		try:
			asynchat.async_chat.initiate_send (self)		
			try: is_working = self.producer_fifo.working ()
			except AttributeError: 	is_working = len (self.producer_fifo)
		finally:	
			lock.release ()			
		if not is_working:
			self.done_request ()		
		
	def initiate_send (self):
		asynchat.async_chat.initiate_send (self)		
		try: is_working = self.producer_fifo.working ()
		except AttributeError: 	is_working = len (self.producer_fifo)
		if not is_working:
			self.done_request ()		
			
	def readable (self):		
		return not self.is_rejected and asynchat.async_chat.readable (self)
		
	def issent (self):
		return self.bytes_out.as_long ()
		
	def __repr__ (self):
		ar = asynchat.async_chat.__repr__(self) [1:-1]
		return '<%s channel#: %s requests:%s>' % (
				ar,
				self.channel_number,
				self.request_counter
				)
		
	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == 3:
			self.reject ()
			if self.writable ():
				return 1
			self.close ()
			return 0
	      		
	def isconnected (self):
		return self.connected
	
	def handle_timeout (self):
		self.log ("killing zombie channel %s" % ":".join (map (str, self.addr)))
		self.close ()
	
	def set_timeout (self, timeout):
		self.zombie_timeout = timeout
	
	def set_socket_timeout (self, timeout):	
		self.keep_alive = timeout
		self.network_timeout = timeout
			
	def attend_to (self, thing):
		if not thing: return
		self.producers_attend_to.append (thing)
	
	def die_with (self, thing, tag):
		if not thing: return
		self.things_die_with.append ((thing, tag))
		
	def done_request (self):	
		self.producers_attend_to = [] # all producers are finished	
		self.set_timeout (self.keep_alive)		
							
	def send (self, data):
		#print	("SEND", repr (data), self.get_terminator ())
		self.event_time = int (time.time())
		result = asynchat.async_chat.send (self, data)		
		self.server.bytes_out.inc (result)
		self.bytes_out.inc (result)
		return result
	
	def recv (self, buffer_size):
		self.event_time = int (time.time())		
		try:
			result = asynchat.async_chat.recv (self, buffer_size)
			lr = len (result)
			self.server.bytes_in.inc (lr)
			self.bytes_in.inc (lr)
			if not result:
				self.handle_close ()
				return b""		
			#print	("RECV", repr(result), self.get_terminator ())
			return result
			
		except MemoryError:
			lifetime.shutdown (1, 1.0)
				
	def collect_incoming_data (self, data):
		#print ("collect_incoming_data", repr (data [:180]), self.current_request)
		if self.current_request:
			self.current_request.collect_incoming_data (data)
		else:
			self.in_buffer = self.in_buffer + data
				
	def found_terminator (self):		
		if self.is_rejected:
			return
			
		if self.current_request:			
			self.current_request.found_terminator()
			if DEBUG:
				self.__history.append ("REQUEST DONE")
				self.__history = self.__history [-30:]
			
		else:
			header = self.in_buffer
			#print ("####### CLIENT => SKITAI ##########################")
			#print (header)
			#print ("------------------------------------------------")
			self.in_buffer = b''			
			lines = header.decode ("utf8").split('\r\n')
			while lines and not lines[0]:
				lines = lines[1:]

			if not lines:
				self.close_when_done()
				return
			
			request = lines[0]
			try:
				command, uri, version = http_util.crack_request (request)
			except:
				self.log_info ("channel-%s invaild request header" % self.channel_number, "fail")
				return self.close ()

			if DEBUG: 
				self.__history.append ("START REQUEST: %s/%s %s" % (command, version, uri))				
			
			header = http_util.join_headers (lines[1:])
			r = http_request.http_request (self, request, command, uri, version, header)
			self.set_timeout (self.network_timeout)
			
			self.request_counter.inc()
			self.server.total_requests.inc()
			
			if command is None:
				r.response.error (400)
				return
			
			for h in self.server.handlers:
				if h.match (r):
					try:
						self.current_request = r
						h.handle_request (r)
												
					except:
						self.server.trace()
						try: r.response.error (500)
						except: pass
							
					return
					
			try: r.response.error (404)
			except: pass
	
	def handle_abort (self):
		self.close (ignore_die_partner = True)
		self.log_info ("channel-%s aborted" % self.channel_number, "info")
			      				
	def close (self, ignore_die_partner = False):
		if self.closed:
			return
		
		for closable, tag in self.things_die_with:
			if closable and hasattr (closable, "channel"):
				closable.channel = None
			self.journal (tag)
			if not ignore_die_partner:
				self.producers_attend_to.append (closable)
				
		if self.current_request is not None:
			self.producers_attend_to.append (self.current_request.collector)
			self.producers_attend_to.append (self.current_request.producer)
			# 1. close forcely or by error, make sure that channel is None
			# 2. by close_when_done ()
			self.current_request.channel = None
			self.current_request = None
			
		for closable in self.producers_attend_to:
			if closable and hasattr (closable, "close"):
				try:
					closable.close ()
				except:
					self.server.trace()
		
		self.producers_attend_to, self.things_die_with = [], []
		self.discard_buffers ()
		asynchat.async_chat.close (self)
		self.connected = False		
		self.closed = True
		self.log_info ("channel-%s closed" % self.channel_number, "info")
	
	def journal (self, reporter):
		self.log (
			"%s closed, client %s:%s, bytes in: %s, bytes out: %s for %d seconds " % (
				reporter,
				self.addr [0], 
				self.addr [1], 
				self.bytes_in,
				self.bytes_out,
				time.time () - self.creation_time
			)
		)
					
	def log (self, message, type = "info"):
		self.server.log (message, type)
	
	def trace (self, id = None):
		self.server.trace (id)
			
	def log_info (self, message, type='info'):
		self.server.log (message, type)
		
	def handle_expt(self):
		self.log_info ("channel-%s panic" % self.channel_number, "fail")
		self.close ()
			
	def handle_error (self):
		self.server.trace ("channel-%s" % self.channel_number)		
		self.close()
	
		
#-------------------------------------------------------------------
# server class
#-------------------------------------------------------------------
class http_server (asyncore.dispatcher):
	SERVER_IDENT = skitai.NAME
	
	maintern_interval = 0
	critical_point_cpu_overload = 90.0
	critical_point_continuous = 3
	
	def __init__ (self, ip, port, server_logger = None, request_logger = None):
		global PID
		
		self.handlers = []
		self.ip = ip
		self.port = port
		asyncore.dispatcher.__init__ (self)

		if ip.find (":") != -1:
			self.create_socket (socket.AF_INET6, socket.SOCK_STREAM)		
		else:			
			self.create_socket (socket.AF_INET, socket.SOCK_STREAM)		

		self.set_reuse_addr ()
		try:
			self.bind ((ip, port))
		except OSError as why:			
			if why.errno == 98:
				server_logger ("address already in use, cannot start server", "fatal")
			else:
				server_logger.trace ()	
			sys.exit (0)
		
		self.worker_ident = "master"
		self.server_logger = server_logger
		self.request_logger = request_logger
		self.start_time = time.ctime(time.time())
		self.start_time_int = time.time()
		
		self.total_clients = counter.mpcounter()
		self.total_requests = counter.mpcounter()
		self.exceptions = counter.mpcounter()
		self.bytes_out = counter.mpcounter()
		self.bytes_in  = counter.mpcounter()
		self.shutdown_phase = 2		
		
		self.__last_maintern = time.time ()
				
		host, port = self.socket.getsockname()
		hostname = socket.gethostname()
		try:			
			self.server_name = socket.gethostbyaddr (hostname)[0]
		except socket.error:			
			self.server_name = hostname
		self.hash_id = md5 (self.server_name.encode ('utf8')).hexdigest() [:4]		
		self.server_port = port
		
	def _serve (self, shutdown_phase = 2):
		self.shutdown_phase = shutdown_phase
		self.listen (os.name == "posix" and 4096 or 256)
	
	def serve (self, sub_server = None): 
		if sub_server:
			sub_server._serve (max (1, self.shutdown_phase - 1))
		self._serve ()			
		
	def fork (self, numworker = 1):
		global ACTIVE_WORKERS, SURVAIL, PID, EXITCODE
		
		if os.name == "nt":
			set_process_name (skitai.get_proc_title ())
			signal.signal(signal.SIGTERM, hTERMWORKER)			
		
		else:			
			while SURVAIL:
				try:	
					if ACTIVE_WORKERS < numworker:						
						pid = os.fork ()
						if pid == 0:							
							self.worker_ident = "worker #%d" % len (PID)
							set_process_name ("%s: %s" % (skitai.get_proc_title (), self.worker_ident))
							PID = {}
							signal.signal(signal.SIGTERM, hTERMWORKER)
							signal.signal(signal.SIGQUIT, hQUITWORKER)
							break
							
						else:
							set_process_name ("%s: master" % skitai.get_proc_title ())
							if not PID:								
								signal.signal(signal.SIGHUP, hHUPMASTER)
								signal.signal(signal.SIGTERM, hTERMMASTER)
								signal.signal(signal.SIGINT, hTERMMASTER)
								signal.signal(signal.SIGQUIT, hQUITMASTER)
								signal.signal (signal.SIGCHLD, hCHLD)
							
							ps = psutil.Process (pid)
							ps.x_overloads = 0
							PID [pid] = ps
							ACTIVE_WORKERS += 1
							#print ('-----', PID, ACTIVE_WORKERS)
					
					now = time.time ()
					if self.maintern_interval and now - self.__last_maintern > self.maintern_interval:						
						self.maintern (now)		
					time.sleep (1)
					
				except KeyboardInterrupt:
					EXITCODE = 0
					DO_SHUTDOWN (signal.SIGTERM)
					break
			
			if self.worker_ident == "master":
				return EXITCODE
				
		self.log_info ('%s (%s) started on %s:%d' % (
			self.SERVER_IDENT, self.worker_ident, self.server_name, self.port)
		)
	
	usages = []
	cpu_stat = []
	def maintern (self, now):
		global PID, CPUs
		
		self.__last_maintern = now
		usages = []
		for ps in PID.values ():
			if ps is None:
				continue
							
			try:
				usage = ps.cpu_percent ()
			except (psutil.NoSuchProcess, AttributeError):
				# process changed, maybe next time
				usage = 0.0
			usages.append ((ps, usage))
		
		self.usages.append (sum ([x [1] for x in usages]) / len (usages))
		self.usages = self.usages [-45:]
		
		self.cpu_stat = []
		max_usage = 0
		for c in (3, 15, 45):
			# 1m, 5m, 15
			array = self.usages [-c:]
			self.cpu_stat.append ((max (array), int (sum (array) / c)))					
		
		if self.cpu_stat [0][0] > 10:
			# 60% for 1m
			self.log ("CPU percent: " + ", ".join (["%d/%d" % unit for unit in self.cpu_stat]))
			
		for ps, usage in usages:						
			if usage < self.critical_point_cpu_overload:
				ps.x_overloads = 0 #reset count
				continue
							
			ps.x_overloads += 1	
			if ps.x_overloads > self.critical_point_continuous:				
				self.log ("process %d is overloading, try to kill..." % ps.pid, 'fatal')
				sig = ps.x_overloads > (self.critical_point_continuous + 2) and signal.SIGKILL or signal.SIGTERM
				try:
					os.kill (ps.pid, sig)
				except OSError:
					pass
			
	def create_socket(self, family, type):
		if hasattr (socket, "_no_timeoutsocket"):
			sock_class = socket._no_timeoutsocket
		else:
			sock_class = socket.socket

		self.family_and_type = family, type
		sock = sock_class (family, type)
		sock.setblocking(0)
		self.set_socket(sock)
	
	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == self.shutdown_phase:
			self.log_info ('shutting down web server: %s' % self.server_name)
			self.close ()		
	
	def close (self):		
		asyncore.dispatcher.close (self)
		for h in self.handlers:
			if hasattr (h, "close"):
				h.close ()				
					
	def writable (self):
		return 0

	def install_handler (self, handler, back = 1):
		if back:
			self.handlers.append (handler)
		else:
			self.handlers.insert (0, handler)

	def remove_handler (self, handler):
		self.handlers.remove (handler)
	
	def log (self, message, type = "info"):
		if self.server_logger:
			self.server_logger.log (message, type)
		else:
			sys.stdout.write ('log: [%s] %s\n' % (type,str (message)))	
	
	def log_request (self, message):
		if self.request_logger:
			self.request_logger.log (message, "")
		else:
			sys.stdout.write ('%s\n' % message)
	
	def log_info(self, message, type='info'):
		self.log (message, type)
	
	def trace (self, id = None):
		self.exceptions.inc()
		if self.server_logger:
			self.server_logger.trace (id)
		else:
			asyncore.dispatcher.handle_error (self)
	
	def handle_read (self):
		pass

	def readable (self):
		return self.accepting
	
	def handle_error (self):
		self.trace()		
		
	def handle_connect (self):
		pass

	def handle_accept (self):
		#print ('-----', self.worker_ident)
		self.total_clients.inc()
		try:
			conn, addr = self.accept()		
		except socket.error:
			self.log_info ('server accept() threw an exception', 'warn')
			return
		except TypeError:
			if os.name == "nt":
				self.log_info ('server accept() threw EWOULDBLOCK', 'warn')
			return
		#self.log_info ('client %s:%d accepted by %s' % (addr [0], addr [1], self.worker_ident))
		http_channel (self, conn, addr)
		
	def handle_expt (self):
		self.log_info ('socket panic', 'warning')
	
	def handle_close (self):
		self.log_info('server shutdown', 'warning')
		self.close()
	
	def status (self):
		global PID
		
		return 	{
			"child_pids": list (PID.keys ()),
			"ident": "%s for %s" % (self.worker_ident, self.SERVER_IDENT),
			'server_name': self.server_name,	
			"start_time": self.start_time, 		
			'hash_id': self.hash_id,	
			"port": self.port,
			"total_clients": self.total_clients.as_long(),
			"total_request": self.total_requests.as_long(), 
			"total_exceptions": self.exceptions.as_long(), 
			"bytes_out": self.bytes_out.as_long(), 
			"bytes_in": self.bytes_in.as_long(),
			"cpu_percent":	self.cpu_stat
		}
			
def hCHLD (signum, frame):
	global ACTIVE_WORKERS, PID
	
	ACTIVE_WORKERS -= 1
	try:
		pid, status = os.waitpid (0, 0)
	except ChildProcessError:
		pass
	else:
		PID [pid]	= None

def hTERMWORKER (signum, frame):			
	lifetime.shutdown (0, 1.0)

def hQUITWORKER (signum, frame):			
	lifetime.shutdown (0, 30.0)
	
def DO_SHUTDOWN (sig):
	global PID
	
	signal.signal (signal.SIGCHLD, signal.SIG_IGN)
	for pid in PID:
		if PID [pid] is None:
			continue			
		try: os.kill (pid, sig)
		except OSError: pass
			
def hTERMMASTER (signum, frame):
	global EXITCODE, SURVAIL
	SURVAIL = False
	EXITCODE = 0
	DO_SHUTDOWN (signal.SIGTERM)

def hQUITMASTER (signum, frame):
	global EXITCODE
	EXITCODE = 0
	DO_SHUTDOWN (signal.SIGQUIT)

def hHUPMASTER (signum, frame):		
	global EXITCODE
	EXITCODE = 3
	DO_SHUTDOWN (signal.SIGTERM)

def configure (name, network_timeout = 0, keep_alive = 0):
	from . import https_server
	http_server.SERVER_IDENT = name
	https_server.https_server.SERVER_IDENT = name + " (SSL)"
	http_channel.keep_alive = https_server.https_channel.keep_alive = not keep_alive and 2 or keep_alive
	http_channel.network_timeout = https_server.https_channel.network_timeout = not network_timeout and 30 or network_timeout
	
		
if __name__ == "__main__":
	pass
	


# pytest framework ---------------------------------------------
import requests
from rs4 import siesta, webtest
from rs4.psutil import Puppet, processutil
import subprocess
import time
import sys
import os
import xmlrpc.client 
    
class Launcher (webtest.Target):
    def __init__ (self, script, port, ssl = False, silent = True, dry = False, **kargs):
        self.__script = script
        self.__port = port
        
        self.__start_opts = ["---port", str (port)]
        for k, v in kargs:
            self.__start_opts.append ("---{}".format (k))
            if v:
                self.__start_opts.append (str (v))
            
        endpoint = "http{}://127.0.0.1".format (ssl and "s" or "")
        endpoint += ":{}".format (port)
        webtest.Target.__init__ (self, endpoint)
        self.__silent = silent
        self.__dry = dry
        self.__servicing = False
        self.__p = Puppet (communicate = False)        
        self.__closed = True
        self.__start ()
            
    def __start (self):
        if self.__dry:
            return
        
        if self.__silent:
            if self.__is_running ():
                self.__servicing = True
            else:    
                self.__p.start ([sys.executable, self.__script, "-d"] + self.__start_opts)
                self.__wait_until ("running")
        else:
            self.__p.start ([sys.executable, self.__script]  + self.__start_opts)
            time.sleep (3)
        self.__closed = False
        
    def __is_running (self):
        command = "{} {} status".format (sys.executable, self.__script)
        res = subprocess.run (command, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)    
        out, err = res.stdout.decode ("utf8"), res.stderr.decode ("utf8")
        return out.find ("running") != -1

    def __wait_until (self, status, timeout = 10):
        for i in range (timeout):
            command = "{} {} status".format (sys.executable, self.__script)
            res = subprocess.run (command, shell = True, stdout = subprocess.PIPE, stderr = subprocess.PIPE)    
            out, err = res.stdout.decode ("utf8"), res.stderr.decode ("utf8")
            if out.find (status) != -1:
                return
            time.sleep (1)
    
    def _close (self):
        if self.__closed:
            return        
        if self.__silent and not self.__servicing:
            self.__p.start ([sys.executable, self.__script, "stop"])
            self.__wait_until ("stopped")
        else:
            self.__p.kill ()
            time.sleep (3)
        self.__closed = True
    stop = _close
        
launch = Launcher
    


import os, sys
import threading
from skitai import DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB
from ..corequest.httpbase import cluster_manager, task
from ..corequest.dbi import cluster_manager as dcluster_manager, task as dtask
from sqlphile import Template

def is_main_thread ():    
    return isinstance (threading.currentThread (), threading._MainThread)

class Command:
    def __init__ (self, name, callback):
        self.name = name
        self.callback = callback
    
    def __call__ (self, *args, **kargs):
        return self.callback (self.name, args, kargs)
    
    def lb (self, *args, **kargs):
        return self.callback (self.name + ".lb", args, kargs)
    
    def map (self, *args, **kargs):
        return self.callback (self.name + ".map", args, kargs)

    
class AsyncService:    
    ASYNCDBA = {
        "asyncon", "backend", "db", "postgresql", "sqlite3", "redis", "mongodb"
    }    
    METHODS = {
        "options", "trace", "upload", "get", "delete", "post", "put", "patch",        
        "rpc", "xmlrpc", "jsonrpc", "grpc", 
        "ws", "wss"
    }.union (ASYNCDBA)
    DEFAULT_REQUEST_TYPE = ("application/json", "application/json")
    DEFAULT_SQL_TEMPLATE_ENGINES = {
        DB_PGSQL: Template (DB_PGSQL),
        DB_SQLITE3: Template (DB_SQLITE3),
    }
    def __init__ (self, enable_requests = True):
        if enable_requests:
            for method in self.METHODS:
                setattr (self, method, Command (method, self._call))
    
    @classmethod
    def add_cluster (cls, clustertype, clustername, clusterlist, ssl = 0, access = [], max_conns = 100):
        if clustertype and clustertype [0] == "*":
            clustertype = clustertype [1:]
        ssl = 0
        if ssl in (1, True, "1", "yes") or clustertype in ("https", "wss", "grpcs", "rpcs"):
            ssl = 1
        if type (clusterlist) is str:
            clusterlist = [clusterlist]    

        if clustertype and "*" + clustertype in (DB_PGSQL, DB_SQLITE3, DB_REDIS, DB_MONGODB):
            cluster = dcluster_manager.ClusterManager (clustername, clusterlist, "*" + clustertype, access, max_conns, cls.logger.get ("server"))
            cls.clusters_for_distcall [clustername] = dtask.TaskCreator (cluster, cls.logger.get ("server"))
        else:
            cluster = cluster_manager.ClusterManager (clustername, clusterlist, ssl, access, max_conns, cls.logger.get ("server"))
            cls.clusters_for_distcall [clustername] = task.TaskCreator (cluster, cls.logger.get ("server"), cls.cachefs)
        cls.clusters [clustername] = cluster
                    
    def __detect_cluster (self, clustername):
        try: 
            clustername, uri = clustername.split ("/", 1)
        except ValueError:
            clustername, uri = clustername, ""
        if clustername [0] == "@":
            clustername = clustername [1:]
        
        try: 
            return self.clusters_for_distcall ["{}:{}".format (clustername, self.app.app_name)], "/" + uri            
        except (KeyError, AttributeError):
            return self.clusters_for_distcall [clustername], "/" + uri
        
    def _call (self, method, args, karg):        
        uri = None
        if args:
            uri = args [0]
        elif karg:
            uri = karg.get ("uri", "")
        if not uri:
            raise AssertionError ("Missing param uri or cluster name")

        try: 
            command, fn = method.split (".")
        except ValueError: 
            command = method
            if uri [0] == "@": 
                fn = "lb"
            else:
                fn = (command in self.ASYNCDBA and "db" or "rest")

        if fn == "map" and not hasattr (self, "threads"):
            raise AttributeError ("Cannot use Map-Reduce with Single Thread")
                
        if command in self.ASYNCDBA:
            return getattr (self, "_a" + fn) ("*" + command, *args, **karg)                    
        else:    
            return getattr (self, "_" + fn) (command, *args, **karg)
        
    def _create_rest_call (self, cluster, *args, **kargs):
        return cluster.Server (*args, **kargs)

    def _use_cache (self, use_cache, rm_cache):
        if rm_cache:
            self.setlu (rm_cache)
            return False
        if isinstance (use_cache, str):
            return self.getlu (use_cache)
        return use_cache

    def _rest (self, method, uri, data = None, auth = None, headers = None, meta = None, use_cache = True, rm_cache = None, filter = None, callback = None, timeout = 10, caller = None):
        return self._create_rest_call (self.clusters_for_distcall ["__socketpool__"], uri, data, method, self.rebuild_header (headers, method, data, False), auth, meta, self._use_cache (use_cache, rm_cache), False, filter, callback, timeout, caller)
    
    def _crest (self, mapreduce = False, method = None, uri = None, data = None, auth = None, headers = None, meta = None, use_cache = True, rm_cache = None, filter = None, callback = None, timeout = 10, caller = None):
        cluster, uri = self.__detect_cluster (uri)
        return self._create_rest_call (cluster, uri, data, method, self.rebuild_header (headers, method, data), auth, meta, self._use_cache (use_cache, rm_cache), mapreduce, filter, callback, timeout, caller)
                
    def _lb (self, *args, **karg):
        return self._crest (False, *args, **karg)
        
    def _map (self, *args, **karg):
        return self._crest (True, *args, **karg)
    
    def _bind_sqlphile (self, dbo, dbtype):
        try:
            template_engine = self.app.get_sql_template () # Atila has
        except AttributeError:
            template_engine = None
        template_engine = template_engine or self.DEFAULT_SQL_TEMPLATE_ENGINES [dbtype]
        return template_engine.new (dbo)
    
    def _create_dbo (self, cluster, *args, **kargs):
        return cluster.Server (*args, **kargs)
            
    def _ddb (self, server, dbname = "", auth = None, dbtype = DB_PGSQL, meta = None, use_cache = False, rm_cache = None, filter = None, callback = None, timeout = 10, caller = None):
        dbo = self._create_dbo (self.clusters_for_distcall ["__dbpool__"], server, dbname, auth, dbtype, meta, self._use_cache (use_cache, rm_cache), False, filter, callback, timeout, caller)
        if dbtype in (DB_PGSQL, DB_SQLITE3):
            return self._bind_sqlphile (dbo, dbtype)
        return dbo
    
    def _cddb (self, mapreduce = False, clustername = None, meta = None, use_cache = False, rm_cache = None, filter = None, callback = None, timeout = 10, caller = None):
        cluster = self.__detect_cluster (clustername) [0]        
        dbo = self._create_dbo (cluster, None, None, None, None, meta, self._use_cache (use_cache, rm_cache), mapreduce, filter, callback, timeout, caller)
        if cluster.cluster.dbtype in (DB_PGSQL, DB_SQLITE3):
            return self._bind_sqlphile (dbo, cluster.cluster.dbtype)
        return dbo    
    
    def _dlb (self, *args, **karg):
        return self._cddb (False, *args, **karg)
    
    def _dmap (self, *args, **karg):
        return self._cddb (True, *args, **karg)
    
    def _adb (self, dbtype, server, dbname = "", auth = None, meta = None, use_cache = False, rm_cache = None, filter = None, callback = None, timeout = 10, caller = None):
        return self._ddb (server, dbname, auth, dbtype, meta, use_cache, rm_cache, filter, callback, timeout, caller)
    
    def _alb (self, dbtype, *args, **karg):
        return self._cddb (False, *args, **karg)
    
    def _amap (self, dbtype, *args, **karg):
        return self._cddb (True, *args, **karg)
    
    def transaction (self, clustername, auto_putback = True):
        cluster = self.__detect_cluster (clustername) [0]
        return cluster.getconn (auto_putback)
    